#!/usr/bin/env perl

use strict;
use warnings;

use FindBin;
use Getopt::Long qw(:config no_ignore_case bundling pass_through);
use File::Basename;

my $usage = <<__EOUSAGE__;

###################################################################################
#
# --create                      create a new Trinotate sqlite database
#
# --embl_dat <string>           EMBL dat file  (ex. swissprot.dat or trembl.dat)
#
# --pfam <string>               Pfam_A.hmm file
#
# --eggnog <string>             concatenated eggnog file
#
# --go_obo <string>             gene ontology obo file
#
##################################################################################

__EOUSAGE__

    ;



my $embl_dat_file;
my $pfam_file;
my $eggnog_file;
my $go_obo_file;
my $help_flag;
my $create_flag;

&GetOptions('h' => \$help_flag,
            'create' => \$create_flag,
            'embl_dat=s' => \$embl_dat_file,
            'pfam=s' => \$pfam_file,
            'eggnog=s' => \$eggnog_file,
            'go_obo=s' => \$go_obo_file,
    );


if ($help_flag) {
    die $usage;
}
unless ($create_flag || $embl_dat_file || $pfam_file || $eggnog_file || $go_obo_file) {
    die $usage;
}


my $bindir = $FindBin::Bin;

if ($create_flag) {
    my $cmd = "java -cp $bindir:$bindir/../java/sqlitejdbc-v056.jar CreateJavaSQLliteTables";
    &process_cmd($cmd);
}

if ($embl_dat_file) {
    # parse uniprot dat file
    my $cmd = "java -cp $bindir:$bindir/../java/sqlitejdbc-v056.jar EMBLParsewithautocommitt $embl_dat_file";
    &process_cmd($cmd);

    # get unique taxonomy listing
    $cmd = "cat Taxonomyvalue.pre | sort -u > TaxonomyUniq";
    &process_cmd($cmd);

    # import taxonomy data
    $cmd = "java -cp $bindir:$bindir/../java/sqlitejdbc-v056.jar TaxonomytoMySQLLite TaxonomyUniq counter";
    &process_cmd($cmd);
}

if ($pfam_file) {
    # import PFAM
    my $cmd = "java -cp $bindir:$bindir/../java/sqlitejdbc-v056.jar HMMDatabaseParse $pfam_file";
    &process_cmd($cmd);
}

if ($eggnog_file) {
    # import EggNog
    my $cmd = "java -cp $bindir:$bindir/../java/sqlitejdbc-v056.jar eggNOGtoMYSQLlite $eggnog_file";
    &process_cmd($cmd);
}

# import gene ontology
if ($go_obo_file) {
    my $go_obo_tab = basename($go_obo_file) . ".tab";
    if (-s $go_obo_tab) {
        print STDERR "-reusing existing $go_obo_tab file\n";
    }
    else {
        my $cmd = "$bindir/../util/gene_ontology/obo_to_tab.pl $go_obo_file > $go_obo_tab";
        &process_cmd($cmd);
    }
    
    my $cmd = "$bindir/../util/gene_ontology/obo_tab_to_sqlite_db.pl $go_obo_tab";
    &process_cmd($cmd);
    
}

exit(0);

####
sub process_cmd {
    my ($cmd) = @_;

    print STDERR "CMD: $cmd\n";
    
    my $ret = system($cmd);
    if ($ret) {
        die "Error, cmd: $cmd died with ret $ret";
    }
    
    return;
}


