#!/usr/bin/env perl

use strict;
use warnings;

use FindBin;
use lib ("$FindBin::Bin/../PerlLib");

use Sqlite_connect;

use Getopt::Long qw(:config no_ignore_case bundling pass_through);

my $usage = <<__EOUSAGE__;

##################################################################
#
#  -E <float>                 maximum E-value for reporting best blast hit
#                             and associated annotations.
#
#  --pfam_cutoff <string>     'DNC' : domain noise cutoff (default)
#                             'DGC' : domain gathering cutoff
#                             'DTC' : domain trusted cutoff
#                             'SNC' : sequence noise cutoff
#                             'SGC' : sequence gathering cutoff
#                             'STC' : sequence trusted cutoff
#
##################################################################

__EOUSAGE__

    ;

my $Evalue_cutoff = 1e-5;
my $help_flag;
my $pfam_cutoff = "DNC";

&GetOptions( 
    'h' => \$help_flag,
    'E=f' => \$Evalue_cutoff,
    'pfam_cutoff=s' => \$pfam_cutoff,
    );

if ($help_flag) {
    die $usage;
}

unless ($pfam_cutoff =~ /^(DNC|DGC|DTC|SNC|SGC|STC)$/) {
    die $usage . "\n\nError, do not recognize pfam_cutoff specified as [$pfam_cutoff]\n\n";
}

main: {

    unless (-s "TrinityFunctional.db") {
        die "Error, must first build Trinotate sqlite results database";
    }
    
    my $dbproc = &connect_to_db("TrinityFunctional.db");

    my @results = &do_sql_2D($dbproc, "select TrinityAminoAcidID, Path, AminoAcid_sequence from TrinityTranscript");

    # print header
    print join("\t", "#component", "trans_derived", "prot_id", "TopBlastHit", "Pfam", "SignalP", "TmHMM", "eggnog", "gene_ontology", "prot_seq") . "\n";
    
    foreach my $result (@results) {
        
        my ($prot_id, $trans_id, $prot_seq) = @$result;

        my $blast_info = &get_blast_results($dbproc, $prot_id);
        my $pfam_info = &get_pfam_info($dbproc, $prot_id);
        
        my $signalP_info = &get_signalP_info($dbproc, $prot_id);
        
        my $TmHMM_info = &get_TmHMM_info($dbproc, $prot_id);

        my $eggnog = &get_eggnog_info_from_blast_hit($dbproc, $blast_info);
        
        my $gene_ontology = &get_gene_ontology_from_blast_hit($dbproc, $blast_info);
                

        $trans_id =~ /^(comp\d+_c\d+)/ or die "Error, cannot parse component from $trans_id";
        my $component_id = $1;


        print join("\t", $component_id, $trans_id, $prot_id, $blast_info, $pfam_info, $signalP_info, $TmHMM_info, $eggnog, $gene_ontology, $prot_seq) . "\n";
    
    }
    
    exit(0);


}


####
sub get_TmHMM_info {
    my ($dbproc, $id) = @_;

    my $query = "select Score, PredHel, Topology from tmhmm where queryprotid = ? and PredHel != 'PredHel=0' ";
    my $result = &first_result_sql($dbproc, $query, $id);
    
    if ($result) {
        my $tmhmm_line = join("`", @$result);
        return($tmhmm_line);
    }
    else {
        return(".");
    }
}


####
sub get_signalP_info {
    my ($dbproc, $id) = @_;

    my $query = "select start, end, score, prediction from SignalP where query_prot_id = ?";
    my $result = &first_result_sql($dbproc, $query, $id);
    if ($result) {
        my $sigP_line = "sigP:" . join("`", @$result);
        return($sigP_line);
    }
    else {
        return(".");
    }
}


###
sub get_pfam_info {
    my ($dbproc, $id) = @_;

    my $query = "select pfam_id, HMMERDomain, HMMERTDomainDescription, QueryStartAlign, QueryEndAlign, ThisDomainEvalue "
        . " from HMMERDbase h, PFAMreference p"
        . " where QueryProtID = ? "
        . " and p.pfam_accession = h.pfam_id ";

    if ($pfam_cutoff eq "DNC") {
        $query .= " and h.FullDomainScore >= p.Domain_NoiseCutoff ";
    }
    elsif ($pfam_cutoff eq "DGC") {
        $query .= " and h.FullDomainScore >= p.Domain_GatheringCutOff ";
    }
    elsif ($pfam_cutoff eq "DTC") {
        $query .= " and h.FullDomainScore >= p.Domain_TrustedCutOff ";
    }
    elsif ($pfam_cutoff eq "SNC") {
        $query .= " and h.FullSeqScore >= p.Sequence_NoiseCutoff ";
    }
    elsif ($pfam_cutoff eq "SGC") {
        $query .= " and h.FullSeqScore >= p.Sequence_GatheringCutOff ";
    }
    elsif ($pfam_cutoff eq "STC") {
        $query .= " and h.FullSeqScore >= p.Sequence_TrustedCutOff ";
    }
    
    $query .=  " order by QueryStartAlign ";
    
    my @results = &do_sql_2D($dbproc, $query, $id);

    if (@results) {
        my @encoded_hits;
        foreach my $result (@results) {
            my ($pfam_id, $domain, $domain_descr, $start, $end, $evalue) = @$result;
            
            my $hit = join("^", $pfam_id, $domain, $domain_descr, "$start-$end", "E:$evalue");
            push (@encoded_hits, $hit);
        }
        
        my $result_line = join("`", @encoded_hits);
        
        return($result_line);
    }
    else {
        return(".");
    }
}




####
sub get_blast_results {
    my ($dbproc, $id) = @_;

    my $query = "select FullAccession, UniprotSearchString, QueryStart, QueryEnd, HitStart, HitEnd, PercentIdentity, Evalue "
        . " from BlastDbase where TrinityID = ?"
        . " and Evalue <= $Evalue_cutoff ";
        ;
    
    my $result = &first_result_sql($dbproc, $query, $id);

    if ($result) {
        my ($FullAccession, $UniprotSearchString, $QueryStart, $QueryEnd, $HitStart, $HitEnd, $PercentIdentity, $Evalue) = @$result;
        
        my $taxonomy_string = &get_taxonomy_string($dbproc, $UniprotSearchString) || ".";
        $taxonomy_string =~ s/[\`\s]+/ /g; 
        $taxonomy_string =~ s/^\s+|\s+$//g;
        
        my $description_line = &get_description_line($dbproc, $UniprotSearchString) || ".";
        
        $description_line =~ s/[\`\s]+/ /g; # using backtics as delimiters, and dont want tabs or newlines to disrupt formatting.
        $description_line =~ s/^\s+|\s+$//g;  
        
        ## encode the result
        my $ret_val = join("`", $FullAccession, $UniprotSearchString, "Q:$QueryStart-$QueryEnd,H:$HitStart-$HitEnd", "$PercentIdentity%ID", "E:$Evalue", $description_line, $taxonomy_string);

        return($ret_val);
    }
    else {
        return(".");
    }

}

####
sub get_taxonomy_string {
    my ($dbproc, $uniprot_acc) = @_;


    ## get the link ID
    my $query = "select u.LinkId from UniprotIndex u where u.AttributeType = 'T' and u.Accession = ?";
    my $link_id = &very_first_result_sql($dbproc, $query, $uniprot_acc);

    

    $query = "select t.TaxonomyValue from TaxonomyIndex t "
        . " where t.NCBITaxonomyAccession = ?";

    my $result = &very_first_result_sql($dbproc, $query, $link_id);
    
    return($result);
}

####
sub get_description_line {
    my ($dbproc, $uniprot_acc) = @_;

    my $query = "select LinkId from UniprotIndex u where u.Accession = ? and u.AttributeType = 'D'";
    my $description = &very_first_result_sql($dbproc, $query, $uniprot_acc) || "";
    $description =~ s/^\s+|\s+$//g; # trim leading/trailing ws
    
    return($description);
}

####
sub get_eggnog_info_from_blast_hit {
    my ($dbproc, $blast_info) = @_;

    if ($blast_info eq ".") {
        # no top hit
        return(".");
    }

    my @vals = split(/\`/, $blast_info);
    my $blast_hit_acc = $vals[1];

    my $query = "select eggNOGIndexTerm, eggNOGDescriptionValue "
        . " from eggNOGIndex e, UniprotIndex u "
        . " where u.Accession = ? "
        . " and u.AttributeType = 'E' "
        . " and u.LinkId = e.eggNOGIndexTerm ";

    my @results = &do_sql_2D($dbproc, $query, $blast_hit_acc);

    if (@results) {
        my @eggnogs;
        foreach my $result (@results) {
            my ($eggnog_acc, $eggnog_descr) = @$result;
            my $token = $eggnog_acc;
            if ($eggnog_descr) {
                $eggnog_descr =~ s/\s+/ /g;
                $token .= "^$eggnog_descr";
            }
            push (@eggnogs, $token);
        }
        my $retval = join("`", @eggnogs);
        return($retval);
    }
    else {
        return(".");
    }
}

####
sub get_gene_ontology_from_blast_hit {
    my ($dbproc, $blast_info) = @_;

    if ($blast_info eq ".") {
        # no top hit
        return(".");
    }

    my @vals = split(/\`/, $blast_info);
    my $blast_hit_acc = $vals[1];

    my $query = "select g.id, g.namespace, g.name "
        . " from go g, UniprotIndex u "
        . " where u.Accession = ? "
        . " and u.AttributeType = 'G' "
        . " and u.LinkId = g.id ";

    my @results = &do_sql_2D($dbproc, $query, $blast_hit_acc);
    
    if (@results) {
        my @tokens;
        foreach my $result (@results) {
            my ($go_id, $go_namespace, $go_name) = @$result;
            
            my $token = join("^", $go_id, $go_namespace, $go_name);
            push (@tokens, $token);
        }
        my $retval = join("`", @tokens);

        return($retval);
    }
    else {
        return(".");
    }
    

}

        
